/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj.internal.kernel;

/**
 * A call is an instance of an {@link KProcedure}.
 * 
 * <h1>Subclasses contract</h1>
 * <p>
 * Subclasses must inform when they {@link #started() start} and when they
 * {@link #completed() complete}.
 * 
 * @see KProcedure#newCall(KProcess)
 * 
 * @author Atos Worldline
 */
public class KCall {

	public enum State {

		/**
		 * The {@link KCall} is ready to be started.
		 */
		PENDING,

		/**
		 * The {@link KCall} is being executed.
		 */
		RUNNING,

		COMPLETED,

		/**
		 * The {@link KCall} has been canceled.
		 */
		CANCELED,

		/**
		 * The {@link KCall} has completed but with exception.
		 */
		FAILED,

		/**
		 * The {@link KCall} has completed successfully.
		 */
		SUCCEEDED,

		/**
		 * 
		 */
		DISCARDED;

		public boolean isFinal() {
			return this == FAILED || this == SUCCEEDED || this == DISCARDED;
		}

	}

	/**
	 * The procedure that created this instance.
	 */
	private final KProcedure procedure;

	/**
	 * The process that owns this instance.
	 */
	private final KProcess process;

	/**
	 * KTask level context.
	 * 
	 */
	Object context;

	/**
	 * The values of the {@link KTask#inputParameters}.
	 */
	Object[] inputValues;

	/**
	 * The values of the {@link KTask#outputParameters}.
	 */
	Object[] outputValues;

	Exception exception;

	private State state = State.PENDING;

	/**
	 * id of the procedure call
	 */
	private String id;

	protected KCall(KProcedure procedure, KProcess process) {
		this.procedure = procedure;
		this.process = process;
		this.inputValues = new Object[this.procedure.inputParameters.size()];
		this.outputValues = new Object[this.procedure.outputParameters.size()];
	}

	public void start() {
		this.process.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KCall.this.state == State.PENDING;
			}

			@Override
			protected void fire() {
				KCall.this.setState(State.RUNNING);
			}
		});
	}

	public void complete() {
		this.process.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KCall.this.state == State.RUNNING;
			}

			@Override
			protected void fire() {
				KCall.this.getProcedure().complete(KCall.this);
				KCall.this.setState(State.COMPLETED);
			}
		});
	}

	public void cancel() {
		this.process.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				// TODO Auto-generated method stub
				return false;
			}

			@Override
			protected void fire() {
				// TODO Auto-generated method stub

			}

		});
	}

	void fail() {
		this.process.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KCall.this.state == State.COMPLETED;
			}

			@Override
			protected void fire() {
				KCall.this.setState(State.FAILED);
			}
		});
	}

	void discard() {
		this.process.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KCall.this.state == State.COMPLETED;
			}

			@Override
			protected void fire() {
				KCall.this.setState(State.DISCARDED);
			}
		});
	}

	void succeed() {
		this.process.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KCall.this.state == State.COMPLETED;
			}

			@Override
			protected void fire() {
				KCall.this.procedure.succeed(KCall.this);
				KCall.this.setState(State.SUCCEEDED);
			}
		});
	}

	public Runnable toRunnable() {
		return new Runnable() {
			@Override
			public void run() {
				KCall.this.start();
				KCall.this.complete();
			}
		};
	}

	protected void setState(State state) {
		this.state = state;
	}

	public Object getContext() {
		return context;
	}

	public void setContext(Object context) {
		this.context = context;
	}

	public KProcess getProcess() {
		return process;
	}

	public Object[] getInputValues() {
		return inputValues;
	}

	public Object[] getOutputValues() {
		return outputValues;
	}

	public Exception getException() {
		return exception;
	}

	protected void setException(Exception exception) {
		this.exception = exception;
	}

	State getState() {
		return state;
	}

	public String getId() {
		return id;
	}

	public KProcedure getProcedure() {
		return procedure;
	}

}