package org.parallelj.internal.kernel;

import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.ReentrantLock;

/**
 * {@link KMachine} is an implementation of a state / transition machine.
 * 
 * <ul>
 * <li>Transitions are represented by {@link KTransition}.</li>
 * <li>States are managed by sub classes of {@link KMachine}.</li>
 * </ul>
 * 
 * At runtime, instances of {@link KTransition} are first enqueued (
 * {@link #firingSequence}) when {@link #submit(KTransition)} is called. Then
 * they are {@link KTransition#fire() fired} if they are
 * {@link KTransition#isEnabled()}.
 * 
 * 
 * Note. Regarding multi-threading aspect.
 * <ul>
 * <li>The firing of {@link KTransition} is done by one thread. Furthermore, a
 * transition needs to be completed before firing a new one.</li>
 * <li>The thread that will fire a transition might not be the thread that
 * enqueued it.</li>
 * <li>But it is ensured that the thread will exit from the
 * {@link #submit(KTransition)} call after the dequeue of the
 * {@link KTransition} that it has submitted.</li>
 * </ul>
 * 
 * @author Laurent Legrand
 * 
 */
public class KMachine {

	/**
	 * Lock used to protect methods that triggers a transition.
	 */
	private ReentrantLock lock = new ReentrantLock();

	/**
	 * The queue of {@link KTransition transitions} to be fired if enabled.
	 */
	private final Queue<KTransition> firingSequence = new ConcurrentLinkedQueue<KTransition>();

	/**
	 * Add a transition to the firing sequence.
	 * 
	 * @param transition
	 */
	protected void submit(KTransition transition) {
		this.firingSequence.offer(transition);
		this.fire();
	}

	/**
	 * Fire all transitions in the {@link #firingSequence}.
	 * 
	 * Reetrance is managed by the {@link #lock}.
	 */
	private void fire() {
		// check if the current thread is already firing a transition.
		if (this.lock.isHeldByCurrentThread()) {
			return;
		}

		this.lock.lock();
		try {
			KTransition transition;
			while ((transition = this.firingSequence.poll()) != null) {
				if (transition.isEnabled()) {
					transition.fire();
				}
			}
		} finally {
			this.lock.unlock();
		}
	}

}
