/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj.internal.kernel;

import java.util.HashSet;
import java.util.Set;

/**
 * Represents an instance of a {@link KProgram}.
 * 
 * It cannot be instantiated directly; {@link KProgram#newProcess(Object)} must
 * be used instead.
 * 
 * @author Atos Worldline
 */
public class KProcess extends KMachine {

	/**
	 * Represents the state of a {@link KProcess}.
	 * 
	 * @author Atos Worldline
	 */
	public enum State {

		/**
		 * This is the initial state of a {@link KProcess}.
		 * 
		 * Transition to {@link #RUNNING} when {@link KProcess#start()} is
		 * triggered.
		 * 
		 */
		PENDING,

		/**
		 * The {@link KProcess} is currently RUNNING meaning that some {@link KCall}  are
		 * currently RUNNING or new {@link KCall} will be launched.
		 * 
		 * <p>
		 * Transition to {@link #COMPLETED} when all {@link KCall} are complete.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #ABORTING} when {@link KProcess#abort()} is
		 * triggered.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #TERMINATING} when
		 * {@link KDefinition#terminate()} is triggered.
		 * </p>
		 * 
		 */
		RUNNING,

		ABORTING,

		TERMINATING,

		/**
		 * This is an exit condition of the process. No procedures are running
		 * and no new procedures can be launched.
		 */
		ABORTED,

		/**
		 * This is an exit condition of the process. No procedures are running
		 * and no new procedures can be launched.
		 */
		TERMINATED,

		/**
		 * This is the default exit condition of the process. The process has
		 * completed its work without interruption (cf. {@link KProcess#abort()}
		 * or {@link KProcess#terminate()}).
		 */
		COMPLETED;

		/**
		 * Return <code>true</code> if this state is a final state.
		 * 
		 * @return
		 */
		public boolean isFinal() {
			return this == COMPLETED || this == ABORTED || this == TERMINATED;
		}

	}

	/**
	 * 
	 */
	private final KProgram program;

	/**
	 * KDefinition level context
	 */
	final Object context;

	/**
	 * Marking of all elements.
	 */
	Object[] markings;

	/**
	 * The state of this process.
	 */
	private State state = State.PENDING;

	/**
	 * id of the process
	 */
	private String id;

	/**
	 * The parent id.
	 */
	private String parentId = "0.0.0.0";

	protected KProcess(KProgram program, Object context) {
		this.program = program;
		this.context = context;
	}

	protected void setState(State state) {
		this.state = state;
	}

	boolean isEnabled() {
		return this.verify() && this.verifyLiveness();
	}

	boolean verifyLiveness() {
		return (this.program.liveness.size(this) < this.program.liveness
				.getCapacity());
	}

	void incrementLiveness() {
		this.program.liveness.produce(this);
	}

	void decrementLiveness() {
		this.program.liveness.consume(this);
	}

	boolean isBusy() {
		return this.program.liveness.size(this) != 0;
	}

	/**
	 * Verify a pre condition which is global to the definition. It will be
	 * called prior to try firing a execution (
	 * {@link KTask#tryFiring(KDefinition)}.
	 */
	public boolean verify() {
		return true;
	}

	/**
	 * Start this process.
	 */
	public final void start() {
		this.submit(new KTransition() {
			
			@Override
			protected boolean isEnabled() {
				return KProcess.this.state == State.PENDING;
			}

			@Override
			protected void fire() {
				for (KProcedure procedure: KProcess.this.program.procedures) {
					if (procedure.isEnabled(KProcess.this)) {
						// at least one procedure is enabled
						KProcess.this.setState(State.RUNNING);
						return;
					}
				}

				// no procedure enabled, => completed
				KProcess.this.setState(State.COMPLETED);

			}
		});

	}

	void handleCompletedCall(final KCall call) {
		this.submit(new KTransition() {
			
			@Override
			protected boolean isEnabled() {
				return KProcess.this.state == State.RUNNING
						|| KProcess.this.state == State.ABORTING
						|| KProcess.this.state == State.TERMINATING;
			}

			@Override
			protected void fire() {
				switch (KProcess.this.state) {
				case RUNNING:
					if (call.getException() == null) {
						call.succeed();
					} else {
						if (call.getProcedure().getHandler() != null) {
							call.getProcedure().getHandler()
									.addCallOnError(call);
							call.fail();
						} else {
							switch (KProcess.this.program.policy) {
							case RESUME:
								call.succeed();
								break;
							case TERMINATE:
								call.discard();
								KProcess.this.terminate();
								break;
							case ABORT:
								call.discard();
								KProcess.this.abort();
								break;
							}
						}
					}

					break;
				case TERMINATING:
					if (call.getException() == null) {
						call.succeed();
					} else {
						call.discard();
					}

					break;
				case ABORTING:
					call.discard();
				}
			}
		});
	}

	/**
	 * Fire the enable procedures.
	 * 
	 * @return the procedure calls that are ready to be executed.
	 */
	public final Set<KCall> fire() {
		final Set<KCall> calls = new HashSet<KCall>();
		this.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KProcess.this.state == State.RUNNING;
			}

			@Override
			protected void fire() {
				for (KProcedure procedure : KProcess.this.program.procedures) {
					// check pre condition
					if (!KProcess.this.isEnabled()) {
						break;
					}
					while (KProcess.this.isEnabled()
							&& procedure.isEnabled(KProcess.this)) {
						KCall call = procedure.newCall(KProcess.this);
						// TODO refactor: should be included in newCall
						procedure.enter(call);
						calls.add(call);
					}
				}

				// check completeness
				if (!KProcess.this.isBusy()) {
					KProcess.this.complete();
				}
			}
		});
		return calls;
	}

	private void complete() {
		this.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KProcess.this.state == State.RUNNING
						|| KProcess.this.state == State.ABORTING
						|| KProcess.this.state == State.TERMINATING;
			}

			@Override
			protected void fire() {
				switch (KProcess.this.state) {
				case RUNNING:
					KProcess.this.setState(State.COMPLETED);
					break;
				case ABORTING:
					KProcess.this.setState(State.ABORTED);
					break;
				case TERMINATING:
					KProcess.this.setState(State.TERMINATED);
					break;
				}

			}
		});
	}

	public final void terminate() {
		this.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KProcess.this.state == State.RUNNING;
			}

			@Override
			protected void fire() {
				KProcess.this.setState(State.TERMINATING);
			}
		});
	}

	public final void abort() {
		this.submit(new KTransition() {
			
			@Override
			protected boolean isEnabled() {
				return KProcess.this.state == State.RUNNING;
			}

			@Override
			protected void fire() {
				KProcess.this.setState(State.ABORTING);
			}
		});
	}
	
	public Object getContext() {
		return context;
	}

	public State getState() {
		return state;
	}

	public String getId() {
		return id;
	}

	public KProgram getProgram() {
		return program;
	}

	public String getParentId() {
		return parentId;
	}

	public void setParentId(String parentId) {
		this.parentId = parentId;
	}
}