/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj.internal.kernel;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.atomic.AtomicLong;

import org.parallelj.internal.kernel.KProcess.State;
import org.parallelj.internal.util.KCalls;
import org.parallelj.internal.util.StateEvent;
import org.parallelj.internal.util.StateListener;

/**
 * A processor is responsible to execute a {@link KProcess process} and all its
 * procedure {@link KCall calls}.
 * 
 * @author Atos Worldline
 * 
 */
public class KProcessor extends KMachine implements
		StateListener<KCall, org.parallelj.internal.kernel.KCall.State> {

	/**
	 * The possible state of a {@link KProcessor}
	 * 
	 * 
	 * @author Laurent Legrand
	 * 
	 */
	public enum State {

		/**
		 * Initial state.
		 * 
		 * Waiting for {@link KProcess} to be executed.
		 * 
		 * <p>
		 * Transition to {@link #RUNNING} when
		 * {@link KProcessor#execute(KProcess)} is called.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #SUSPENDED} when {@link KProcessor#suspend()} is
		 * called.
		 * </p>
		 */
		PENDING,

		/**
		 * The {@link KProcessor} is active and is executing {@link KProcess}
		 * and {@link KCall}.
		 * 
		 * <p>
		 * Transition to {@link #PENDING} when all
		 * {@link KProcessor#execute(KProcess)} are in final states.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #SUSPENDED} when {@link KProcessor#suspend()} is
		 * called.
		 * </p>
		 */
		RUNNING,

		/**
		 * The execution of {@link KProcess} and {@link KCall} is suspended.
		 * 
		 * <p>
		 * Transition to {@link #RUNNING} when {@link KProcessor#resume()} is
		 * called.
		 * </p>
		 * 
		 */
		SUSPENDED;

	}

	State state = State.PENDING;

	/**
	 * stores the current processor; if any.
	 * 
	 */
	private static ThreadLocal<KProcessor> current = new ThreadLocal<KProcessor>();

	/**
	 * Sequence for processor ids.
	 */
	private static AtomicLong sequence = new AtomicLong(1);

	/**
	 * Identifier of this process.
	 */
	private long id = sequence.getAndIncrement();

	/**
	 * Executor that will execute the runnable
	 */
	private Executor executor;

	/**
	 * Set of active processes
	 */
	Set<KProcess> processes = new HashSet<KProcess>();

	public KProcessor() {
		this(null);
	}

	public KProcessor(ExecutorService executor) {

		// if the executor is null, create a dummy one that just run the command
		this.executor = (executor != null) ? executor : new Executor() {

			@Override
			public void execute(Runnable command) {
				command.run();
			}
		};
	}

	/**
	 * 
	 * 
	 * @param process
	 */
	public void execute(final KProcess process) {
		this.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return true;
			}

			@Override
			protected void fire() {
				KProcessor.this.processes.add(process);
				KProcessor.this.fire(process);
				if (KProcessor.this.state == State.PENDING) {
					KProcessor.this.state = State.RUNNING;
				}
			}
		});
	}

	public void suspend() {
		this.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KProcessor.this.state != State.SUSPENDED;
			}

			@Override
			protected void fire() {
				KProcessor.this.state = State.SUSPENDED;
			}
		});
	}

	public void resume() {
		this.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KProcessor.this.state == State.SUSPENDED;
			}

			@Override
			protected void fire() {
				for (KProcess process : KProcessor.this.processes) {
					KProcessor.this.fire(process);
				}
				KProcessor.this.state = State.RUNNING;
			}
		});
	}

	void fire(final KProcess process) {
		this.submit(new KTransition() {

			@Override
			protected boolean isEnabled() {
				return KProcessor.this.state == State.RUNNING;
			}

			@Override
			protected void fire() {
				// start the process if not already started
				if (process.getState() == KProcess.State.PENDING) {
					process.start();
				}

				// process is ended ?
				if (process.getState().isFinal()) {
					KProcessor.this.processes.remove(process);
					if (KProcessor.this.processes.size() == 0) {
						KProcessor.this.state = State.PENDING;
					}
					return;
				}
				// submit all procedure calls
				for (KCall call : process.fire()) {
					KCalls.addStateListener(call, KProcessor.this);
					KProcessor.this.submit(call.toRunnable());
				}
			}
		});
	}

	@Override
	public void stateChanged(
			StateEvent<KCall, org.parallelj.internal.kernel.KCall.State> event) {
		if (event.getState().isFinal()) {
			KCalls.removeStateListener(event.getSource(), this);
			this.fire(event.getSource().getProcess());
		}
	}

	/**
	 * Submit a runnable to be executed by the executor
	 * 
	 * @param runnable
	 */
	public void submit(final Runnable runnable) {
		this.executor.execute(new Runnable() {

			@Override
			public void run() {
				try {
					// install this processor in the thread local
					KProcessor.current.set(KProcessor.this);
					runnable.run();
				} finally {
					// uninstall this processor in the thread local
					KProcessor.current.set(null);
				}
			}
		});
	}

	public static KProcessor currentProcessor() {
		return current.get();
	}

	public long getId() {
		return id;
	}
	
	public State getState() {
		return this.state;
	}

}
